/*
                 Copyright(c) 2005 Toufik Sarni

This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA*/

#include <kernel.h>
#include <proc.h>
#include <idt.h> //contient les rferences externes des routines d'interruption

#define adr_idt 0x50000 //@ o se logera la table des vecteurs d'IT
#define taille_idt 256*8

int set_idt(void* offset,short int selecteur,short int type,short int numero_descripteur);
int set_gdt(void* offset,short int selecteur,short int type,short int numero_descripteur);

struct idtr_str{
        short int base;
        unsigned long adr;
        }idtr={taille_idt,adr_idt};

struct idt_str{       /*on initialise le descripteur d'IT  l'@ adr_idt*/
      short int offset0_15; //16bit
      short int selecteur;
      short int type;
      short int offset16_31;
       }*idt=(idt_str *)adr_idt; //idt va pointer sur un descripteur d'IT

void init_idt(){
  print("Initialisation de la IDT ...");

  set_idt(int0,0x8 /*le descripteur 0x38 de la GDT*/,0x8E00/*INTGATE voir doc intel*/,0);
  set_idt(int1,0x8,0x8E00,1);
  set_idt(int2,0x8,0x8E00,2);
  set_idt(int3,0x8,0x8E00,3);
  set_idt(int4,0x8,0x8E00,4);
  set_idt(int5,0x8,0x8E00,5);
  set_idt(int6,0x8,0x8E00,6);
  set_idt(int7,0x8,0x8E00,7);
  set_idt(int8,0x8,0x8E00,8);
  set_idt(int9,0x8,0x8E00,9);
  set_idt(intA,0x8,0x8E00,0xa);
  set_idt(intB,0x8,0x8E00,0xb);
  set_idt(intC,0x8,0x8E00,0xc);
  set_idt(intD,0x8,0x8E00,0xd);
  set_idt(intE,0x8,0x8E00,0xe);
  set_idt(intF,0x8,0x8E00,0xf);
  set_idt(int10,0x8,0x8E00,0x10);
  set_idt(int11,0x8,0x8E00,0x11);


  set_idt(0,SCHED_TACHE,0x8500,0x50);   //task gate: pointe vers la GDT, (descripteur TSS) de la tache 0 (voir "proc.h" pour tache0)
  //set_idt(int50,0x8,0x8E00,0x50);

  set_idt(int51,0x8,0x8E00,0x51);
  set_idt(int52,0x8,0x8E00,0x52);
  set_idt(int57,0x8,0x8E00,0x57);
  set_idt(int60,0x8,0x8E00,0x60);     // IRQ8

  //set_gdt(debug,0x8,0x8C00,10);
  asm{
  lidt fword ptr idtr} //initialisation du registre idtr
  print(" Adresse:0x%X  Taille:%d\n",adr_idt,taille_idt);

}

int set_idt(void *offset,short int selecteur,short int type,short int num)
      {
        unsigned long adr=(unsigned long)offset;
        if (num >=256) return -1;
        (idt+num)->offset0_15= (adr & 0xFFFF); //pour avoir les 16bit du pds faible
        (idt+num)->selecteur=selecteur;
        (idt+num)->type=type;
        (idt+num)->offset16_31=(adr & 0xFFFF0000) >> 0x10; //on fait 1 masque puis on dcale de 16bit vers la droite
      return 0;
      }
int set_gdt(void *offset,short int selecteur,short int type,short int num)
      {
        struct idt_str *gdt=(struct idt_str *)0x40000;
        unsigned long adr=(unsigned long)offset;
        if (num >=256) return -1;
        (gdt+num)->offset0_15= (adr & 0xFFFF); //pour avoir les 16bit du pds faible
        (gdt+num)->selecteur=selecteur;
        (gdt+num)->type=type;
        (gdt+num)->offset16_31=(adr & 0xFFFF0000) >> 0x10; //on fait 1 masque puis on dcale de 16bit vers la droite
      return 0;
      }
