/*
                 Copyright(c) 2005 Toufik Sarni

This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA*/

#include <kernel.h>
#include <memory.h>
#include <proc.h>

#define kr_reserve_pge       2048                  // 8MB=2048*4096
#define kr_reserve_mem       0x800000              // 
#define kr_bin_pge           256                   // (2048/8) octets de pages
#define adr_alloc_t          0x70000               // adresse de la table d'allocation binaire de pages
#define debut_alloc_t        0x70100               // kr_bin_pge+adr_alloc_t
#define max_alloc_pge        1021                  // pas plus de 4MB d'allocation y compris les 2 pages de la pile (1021+2=1023)
#define entree_pgd           num_pgd(adr_process)

u32 fin_alloc_t,                  // adresse de fin de la table d'allocation
    taille_alloc_t,               // taille de la table d'allocation
    taille_mem,                   // taille totale de la mmoire physique.
    libre_alloc_t_div4;           // pour eviter ce calcul  chaque appel de fonction (alloc_page())

struct super_struct_setup *superstruct=(super_struct_setup *)0x10001;

u32 vtr(struct tache *tache,u32 adr)     //fonction qui transforme l'@ virtuelle (adr) en @ relle (renvoye)
 {
   return (((u32)(tache->mem.debut))+(adr-(adr_process)));
 }
u32 svtr(struct tache *tache,u32 adr)   //fonction qui transforme l'@ de la pile virtuelle (adr) en @ relle (renvoye)
 {
   u32 t=align_adr_page(adr_stack);
   if (adr < t)                                  // si on est dans la deuxime page de pile
    return (((u32)tache->mem.stack.debut2)-((adr_stack)-adr));  //debut2 contient l'@ relle de la 2eme page de pile dcale i.e:(exemple: si on a 0x30000 est l'@ de la pile alors debut2=0x30ffe)
   else return (((u32)tache->mem.stack.debut1)-((adr_stack)-adr));

 }


unsigned long mem_libre()
  {
   return (_mem_libre(debut_alloc_t,libre_alloc_t_div4))/1024;
  }

inline void *alloc_page()      //inline pour dire qu'elle va tre souvent sollicite
 {
  return _alloc_page(debut_alloc_t,kr_reserve_mem,libre_alloc_t_div4); // routine assembleur voir: _allocpg.asm
 }

int liberer_page(u32 adr)
{
   u32 p=align_adr_page(adr);
   if (p<=kr_reserve_mem) return 1;
   u8 *ptr=(unsigned char *)adr_alloc_t, octet,decal;

   decal=~(1<<((p/4096)%8));      // avoir un 0  la position du dcalage de la page dans l'octet
                                  // ex: si une page est  la position 3 dans l'octet x, alors decal=xxxx1xxxb
   octet=((p/4096)/8);            // avoir la position de l'octet dans la table d'allocation
   ptr +=octet;
   *ptr &=decal;                 // on met  jour l'octet en question dans la table d'allocation
return 0;
}

unsigned long memtopage(unsigned long taille)
{
  return (taille%4)?((taille/4)+1):(taille/4);

  /* equiv  dire que si le reste de la div de "taille" par 4KO(taille page)
   * n'est pas nul alors retourner la partie entire + une page
   * sinon retourner seulement la partie entire de taille/4
   */
}

void *alloc_pgd(struct tache *tache)
{
paging_off();
   void *p=alloc_page();
   if (p==NULL){
              paging_on();
              return p;
               }
   else
    {
     tache->mem.pgd = (u32 *)p;
     tache->tss.cr3 =(u32)p;

    }
paging_on();
return (void *)p;
}

int liberer_pgd(struct tache *tache)
{
 if (liberer_page((u32)tache->mem.pgd)) return 1;
 return 0;
}
  
void *alloc_pgt(struct tache *tache, unsigned short nbr)
{
paging_off();                // dsactiver la pagination

   u16 i;
   u32 *p=(u32 *)alloc_page(),
                 *ptr=p,*temp=p;
  if (p==(u32 *)NULL){ paging_on();
                return (void *)p;
              }       
   tache->tss.eip=adr_process; //debut de la zone mmoire virtuelle du processus
   if (nbr > max_alloc_pge) nbr=max_alloc_pge;    //pas + de 1022 pages pour un processus

   for(i=1;i<=nbr;i++,ptr++)   //remplir la table des pages
     {
      p=(u32 *)alloc_page();
      if (p==(u32 *)NULL) { tache->mem.pgt=(u32 *)temp;
                     tache->mem.nbr_pgt=(i-1);
                     tache->mem.taille=(i-1)*4;
                     paging_on();
                     return (void*)temp;
                   }
      if (i==1) tache->mem.debut=p;   //avoir le dbut mmoire du processus(premire page) 
      *ptr=((u32)p)|(tache->mem.droit);
     }
   *((tache->mem.pgd)+(entree_pgd))=((u32)temp)|tache->mem.droit;                 //charger l'adresse de la pgt dans
   tache->mem.nbr_pgt=nbr;
   tache->mem.taille=nbr*4;
   tache->mem.pgt=(u32 *)temp;

paging_on();                //ractiver la pagination

  return (void *)temp;
}

//-----------------------------------------

int liberer_pgt(struct tache *tache)
{
 paging_off();
   u32 *ptr_pgt=tache->mem.pgt;
   u8 i;

 for(i=1;i<=tache->mem.nbr_pgt;i++)

                liberer_page((u32)ptr_pgt++);
  
 paging_on();
}


void *alloc_stack(struct tache *tache)
 {
paging_off();
  void *ptr;
                                   // au max 2 pages  allouer pour la pile
  switch (tache->mem.stack.nbr_pge)
    {

      case 2: paging_on();
              return (void*)NULL;   // car le processus en a dja le maximum

            /* par convention, le pointeur de pile est  0xFFFFFFFE
             * pour ce faire, nous allons allouer une page pour la pile, son entree
             * alors dans la table des pages sera la dernire (1023=0x3FF)
             * de mme que pour le rpertoire des pages, d'o alors l'adresse virtuelle 0xFFFFFFFE
             */

      case 1:  //on est dans le cas d'un appel de la part du processus pour tendre ou allouer + d'espace de pile

            ptr=alloc_page();
            if (ptr==(void*)NULL) { paging_on();
                                   (void *)NULL;}

            *((tache->mem.pgt)+1022)=((u32)ptr|tache->mem.droit);  //charger la dernire entre de la table des pages, avec l'adresse de la page alloue
            /*il est inutile dans ce cas, de mettre  jour le rpertoire des pages,
              car il est dj initialis lors de sa cration (voir case 0 ci dessous)*/
            
            tache->mem.stack.debut2=(u32*)(((u32)ptr)+reel_stack);  //reel_stack: voir "proc.h"
            tache->mem.stack.nbr_pge=2;
            paging_on();
            return ptr;

      case 0:                 // on est dans le cas de la cration du precessus, on alloue par dfaut 1 page pour la pile
            ptr=alloc_page();
            if (ptr==(void*)NULL) { paging_on();
                                   return (void *)NULL;}
           
            *((tache->mem.pgt)+1023)=((u32)ptr|tache->mem.droit);  //charger la dernire entre de la table des pages, avec l'adresse de la page alloue
            *((tache->mem.pgd)+1023)=((u32)(tache->mem.pgt)|tache->mem.droit); //charger la dernire entre du rpertoire des pages, avec l'adresse de la table des pages
            tache->mem.stack.debut1=(u32*)(((u32)ptr)+reel_stack);  
            tache->tss.esp=adr_stack;      //adr_stack: voir <proc.h>
            tache->tss.ebp=adr_stack;
            tache->mem.stack.nbr_pge=1;
            paging_on();
            return ptr;
    }        
paging_on();
 }
//----------------------------------------------
int liberer_stack(struct tache *tache)
 {
  if (tache->mem.stack.nbr_pge == 1)
      {
      if (liberer_page((u32)tache->mem.stack.debut1)) return 1;
      else return 0;
      }
  else
     {
      if (liberer_page((u32)tache->mem.stack.debut1)) return 1;
      else
       {
       if (liberer_page((u32)tache->mem.stack.debut2)) return 1;
       else return 0;
       }
     }
 return 0;
 }

void init_table_alloc()
{

 /* on rserve les premires pages pour le noyau
  * 96 octets seront alors alloues quiv  0 --> 300000h => 768 pages
  * puisque la table est binaire (chaque bit reprsente l'tat d'une page)
  * alors 768/8-bit= 96 octets
  */
  unsigned char * ptr=(unsigned char *)adr_alloc_t;

  for (;ptr<(unsigned char *)debut_alloc_t;ptr++)    // de 0x70000  0x70000+96
   { *ptr=0xFF;};                                    // <=> 11111111b le 1 symbolise une page allou (0 si libre)

   unsigned long page=memtopage(taille_mem);               //page= nbr de pages totales prsentes
   fin_alloc_t=((page%8)?((page/8)+1):(page/8))+(adr_alloc_t); //on calcul l'@ de fin de la table d'allocation

   for(ptr=(unsigned char*)debut_alloc_t; ptr<(unsigned char*)fin_alloc_t;ptr++)
           { *ptr=0;};                              //marquer toutes les autres pages comme tant libre

  taille_alloc_t=fin_alloc_t-(adr_alloc_t); // taille de la table
  libre_alloc_t_div4=(fin_alloc_t-(debut_alloc_t))/4;

  print("Table d'allocation de pages:\n   Adresse:0x%X     Taille:%d Octets    Allou:%d Pages.\n",
          adr_alloc_t,taille_alloc_t,kr_reserve_pge);
}


 int init_memory()
  { int i;
    void *p;
    taille_mem=(ram.extd_ram);
    init_table_alloc();

  print("Mmoire libre: %dK\n",mem_libre());
  return 0;
  }
