/*
                 Copyright(c) 2005 Toufik Sarni

This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA*/

#include <kernel.h>
#include <proc.h>
#include <idt.h>

//struct temp_pile_int temp_pile_int;
struct temp_pile_tache temp_pile_tache;
struct tache kernel,scheduler;

struct ldt{
          u32 p_faible; u32 p_fort;
           }ldt;

struct tss_desc{                 //descripteur de tche.
                 u16 limit0_15;
                 u16 base0_15;
                 u8  base16_23;
                 u8  type;
                 u8  gran_lim16_19;
                 u8  base24_31;
               }*tss_desc=(struct tss_desc *)(adr_init_gdt+tache0);

struct tache tache[max_tache];
int set_tss_ldt(void *adr,u32 taille,int num,u8 type)
    {
      if (num > max_tache) return 1;
      if (type==1) type=0x82/*ldt*/; else type=0x89/*TSS*/; 
      (tss_desc+num)->limit0_15=(taille & 0xFFFF);
      (tss_desc+num)->base0_15=(((u32)adr) & 0xFFFF);
      (tss_desc+num)->base16_23=(u8)(((u32)adr) >> 16);
      (tss_desc+num)->type=type;   //Prsent=1,DPL=00,S=0(systeme),type=10B1 (B=0 libre)
      (tss_desc+num)->gran_lim16_19=((taille >>16)& 0x0F); // &0x0f pour G=0(limit en octets), D=0 d'aprs intel, res=0,avl=0 
      (tss_desc+num)->base24_31= (u8)(((u32)adr)>>24);
      return 0;
   }


void *get_tss_adr(u16 index)
   {
     struct tss_desc *tss_struct=(struct tss_desc *)(adr_init_gdt+index);
     if (index > max_tache) return (void *)NULL;

     return (void *)
      ((tss_struct->base0_15) | ((tss_struct->base16_23)<<16) |
         ((tss_struct->base24_31)<<24));
   }

u32 get_tss_taille(u16 index)
  {
     struct tss_desc *tss_struct=(struct tss_desc *)(adr_init_gdt+index);
     if (index > max_tache) return 0;

    return (u32)
       ((tss_struct->limit0_15) | (((tss_struct->gran_lim16_19)&0x0F)<<16));
  }

int get_gdt_libre()  //renvoie un index libre de la GDT, -1 en cas d'erreur
    {
      struct tss_desc *ptr=tss_desc;      
      int i;
      for (i=0;i<max_tache;i++,ptr++)
        {
          if (((ptr->type)&0x80)==0)    // si le bit P=0 (Prsence)
                  return i;           
        };
      return -1;
    }
void rem_gdt_tache(u16 i)
   {
    struct tss_desc *ptr=(tss_desc+i);
    ptr->type &=0x7F;                //mise  0 du bit P (prsence)
    }
int add_gdt_tache()
 {  
    int i=get_gdt_libre();
    if (i!=-1) set_tss_ldt((void *)&(tache[i].tss),taille_tss,i,0);
  return i;
 }

void * creer_processus(char *nom,u32 taille)      // crer un processus en mode USER; N.B: taille est en KO
{   
 
 int i=add_gdt_tache();
 if (i!=-1)                             // si une entree libre existe dans la GDT
  {
    u8 *ptache=(u8 *)&tache[i];
    u16 j=1;
    for(;j<=sizeof(tache[0]);j++,ptache++) *ptache=0;   //initialiser tous les lments de la structure tache[i]  0
                                                        //pour viter toute eventuelle confusion.

    tache[i].mem.droit=7;          // droits sur les pages  allouer,1=USER,1=R/W,1=Prsence
                                   //l'initialistaion du mem.droit doit se faire avant l'appel de "alloc_pgt", car cette dernire l'utilise

   if ((alloc_pgd(&tache[i]) !=(void *)NULL)
     && (alloc_pgt(&tache[i],memtopage(taille))!=(void *)NULL)  // ces 3 fonctions remplissent aussi 
      && (alloc_stack(&tache[i])!=(void *)NULL))               // le infos TSS du processus i
              {   

                 add_file_pret(&tache[i],&file_pret_tpartage);  //ajouter le processus dans la file des processus prets

                 charger_segments;              // macro dans "proc.h"
                 
                 tache[i].tss.tr=((i*8)+tache0);
                 strcpy((char *)&tache[i].nom,nom);
       paging_off();                  //dsactiver la pagination, car on accedera   l'@ pgd du processus, qui est en dehors de l'espace pagin du noyau, voir "init_pge.cpp"

                 u32 *processus_pgd=(u32 *)tache[i].mem.pgd;
                 *processus_pgd=(ADR_KR_PGT1)|5;
                 *(++processus_pgd)=(ADR_KR_PGT2)|5;
                                                 // ajouter au pgd du processus, une entree vers la pgt du noyau,
                                                 // ceci pour pouvoir executer les routines du noyau  partir du contexte utilisateur
                                                 // (ex: en cas d'IT dans l'espace utilisateur, CR3=pgd est celui du processus,
                                                 // alors que le code de la routine d'IT contient l'@ du noyau, en ajoutant l'entree noyau dans le pgd du processus,
                                                 // la routine pourra alors tre xecute
                 *(tache[i].mem.debut)=0xfeebfeeb/*2124*/;
       paging_on();
                 tache[i].tss.link=tache0;
                 tache[i].tss.ss0=KR_DS;     //KR_DS: voir kernel.h et ..\init_gdt.asm
                 tache[i].tss.esp0=(u32)(ADR_TACHE_STACK_IT-((i+1)*1024));  //adresse de la pile, lors d'interruption.
                                                                            // i+1 pour ne pas reserver la premire zone stack i.e: de 0x3ffffe  (0x3ffffe-1Ko)

                 tache[i].tss.eflags=0x200;   //autorise les IT pour les processus
                 tache[i].tss.io_map=104;
                 tache[i].tss.io_per[0]=~0;   //la non permision d'acceder aux 32 premiers ports
                 
                 tache[i].tss.ldt=ldt0;       //ldt0: voir "proc.h"
                 tache[i].id=i;
                 tache[i].etat=pret;
                 return (void *)tache[i].mem.debut;
              }
    else return (void *)NULL;
    }
else return (void *)NULL;     
}
u16 get_pid()
{
  u16 tr;
  asm{str tr};
  return (tr-(tache0))/8;
}

int kill_processus()
  {
   u16 i= get_pid();
   struct tache *ptr=(struct tache *)get_tss_adr(i);
   rem_gdt_tache(i);
   rem_file_pret(ptr,&file_pret_tpartage);
   ptr->etat=termine;
   return 0;
   }

void print_process(struct tache *tache)
{
 print("Processus: %s\nId=%d\n",tache->nom,tache->id);
 print("Dbut mmoire physique: %p\n",tache->mem.debut);
 print("Taille: %dK\n",tache->mem.taille);
 print("Infos hardware:\n");
 print("EAX=%X EBX=%X ECX=%X EDX=%X\n",tache->tss.eax,tache->tss.ebx,
                                       tache->tss.ecx,tache->tss.edx);                             
 print("ESI=%X EDI=%X EBP=%X ESP=%X\n",tache->tss.esi,tache->tss.edi,
                                       tache->tss.ebp,tache->tss.esp);

 print("EIP=%X CR3=%X EFLAGS=%X\n",tache->tss.eip,
                                   tache->tss.cr3,tache->tss.eflags);
 
 print("CS=%x DS=%x ES=%x SS=%x\n",tache->tss.cs,tache->tss.ds,tache->tss.es,
                                   tache->tss.ss);
 print("Etat de la pile:\n");
   paging_off();
    
    u32 *ptr=(u32*)(svtr(tache,tache->tss.esp)), *virt=(u32 *)tache->tss.esp; //pour avoir l'adresse physique
    u16 i;
    for (i=0;i<=16,ptr<=((u32 *)svtr(tache,tache->tss.ebp));i++,ptr++,virt++)  //svtr: macro dans "memory.h" pour transformer l'@ virtuelle(stack) en @ relle
      {
        print("[%p]:%p ptr=%p\n",virt,*ptr,ptr);
      };
   paging_on();
   }

void init_process()
{
  u16 ldtr=ldt0;
  print("\nCration des processus systmes...[ ");

tss_desc=(struct tss_desc *)(adr_init_gdt);  // dclar comme global initialis  (adr_init_gdt+tache0)=0x40000+0x70
                                             // on le fait pointer temporairement sur adr_init_gdt,
                                             // pour faire les initialisations suivantes:
                                             
//mise en place de la LDT par dfaut
  set_tss_ldt((void *)&ldt,7,(ldt0)/8,1);   //1 pour dire: c'est une LDT

//initialisation de KR_TACHE (dscripteur dfinit dans ..\init_gdt.asm)
  set_tss_ldt((void *)&kernel,taille_tss,(KR_TACHE)/8,0); // >>8:pour avoir le numro du dscripteur sgment.
                                                           // 0  pour indiquer que c'est un TSS qu'on dfinit
//initialisation de SCHED_TACHE
  set_tss_ldt((void *)&scheduler,taille_tss,(SCHED_TACHE)/8,0); 
tss_desc=(struct tss_desc *)(adr_init_gdt+tache0);  // on le remet  sa valeur initiale,
                                                    // car il sera utilis  chaque cration d'un processus

    u8 *ptache=(u8 *)&kernel;
    u16 cpt=1;
    for(;cpt<=sizeof(kernel);cpt++,ptache++) *ptache=0;   //initialiser tous les lments de la structure tache[0]  0
                                                            //pour viter toute eventuelle confusion.
     ptache=(u8 *)&scheduler;
     for(cpt=1;cpt<=sizeof(scheduler);cpt++,ptache++) *ptache=0;

     u16 tr=KR_TACHE;
     strcpy(kernel.nom,"Kernel");
     strcpy(scheduler.nom,"Ordonnanceur");

     #include <init_tss.h>  //initialise le contexte de l'ordonnanceur.
     asm{
        ltr tr             // initialise le registre de tche  celui du kernel 
        lldt ldtr          // initialise une LDT par dfaut.
        };
    u32 *gdt_b=(u32 *)(adr_init_gdt+tr);
    *(++gdt_b) &=0xfffffdff;                 // mise  0 du bit Busy dans le descripteur tss de la tche kernel

    u16 sched=SCHED_TACHE;
    printk("OK",i_vert);
    print(" ]\n Kernel: Entre GDT: %xh    Adresse-TSS: %ph  LDT: %xh\n",tr,&kernel,ldtr);
    print(" Scheduler: %xh    TSS: %ph\n",sched,&scheduler);
    
}
